;-------------------------------------------------------------------------
;			UART driver and interrupot handler
;-------------------------------------------------------------------------
;.equ	bauddivider = (xtal / baudrate + 8) / 16 - 1
;-------------------------------------------------------------------------
;				initialization
;-------------------------------------------------------------------------
uart_init:
;	ldi	a0, low(bauddivider)		;set baud rate
;	ldi	a1, high(bauddivider)
;
;input: a1:a0 = baud rate divider
set_baud:
	ldi r30, UBRR0L
	clr r31
	st Z+, a0
	st Z, a1
;	out	UBRR0L, a0
;	out	UBRR0H, a1

	clr	a0				;no U2X, MPCM
	ldi r30, UCSR0A
	clr r31
	st Z+, a0
;	out	UCSR0A, a0

	mov	rx_ptr, xl			;initialize rx buffer

	ldi	a0, 0x01
	ldi	a1, 0xA0
	movw	poly_hi:poly_lo, a1:a0		;set CRC polynom

	ldi	a0, 1<<RXEN0^1<<TXEN0^1<<RXCIE0	;enable RX, TX
	st Z+, a0
;	out	UCSR0B, a0
	ldi	a0, 1<<UCSZ01^1<<UCSZ00	;8 Bit
	st Z+, a0
;	out	UCSR0C, a0
	ret
;-------------------------------------------------------------------------
;				Receive Interrupt
;-------------------------------------------------------------------------
URXCint:
	ldi r30, UDR0
	clr r31
	ld	ia0, Z
	ldi	xh, high( rx_buff )
;	in	ia0, UDR0
	st	X+, ia0
;					calculate CRC
	in	savesreg, sreg
	eor	crc_lo, ia0
	ldi	ia0, 8
_crc1:	lsr	crc_hi
	ror	crc_lo
	brcc	_crc2
	eor	crc_lo, poly_lo
	eor	crc_hi, poly_hi
_crc2:	dec	ia0
	brne	_crc1
	out	sreg, savesreg		;+ 67 cycle
;
	reti				;81 cycle
;-------------------------------------------------------------------------
;				Receive Byte
;-------------------------------------------------------------------------
;output: a0 = byte
;
getchar:				;wait until one byte received
	cp	rx_ptr, xl
	breq	getchar
	mov	yl, rx_ptr
	ldi	yh, high( rx_buff )
	ld	a0, Y
	inc	rx_ptr
	ret
;-------------------------------------------------------------------------
;				Check if command received
;-------------------------------------------------------------------------
;Output: C = 1: command received
;	Y = command start
;	rx_ptr = after command end
;
rx_command:
	mov	yl, rx_ptr
_rxc1:	cp	yl, xl
	breq	_rxc3
	ldi	yh, high( rx_buff )	;wrap around
	ld	a0, Y+
	subi	a0, key_enter
	breq	_rxc2
	subi	a0, key_linefeed - key_enter
	brne	_rxc1
_rxc2:
	mov	a1, yl
	st	-Y, a0			;end of command = 0
	cpse	yl, rx_ptr		;ignore empty command
	sec
	mov	yl, rx_ptr		;point on start of command
	mov	rx_ptr, a1		;point after command
_rxc3:	ret
;-------------------------------------------------------------------------
;				transmit byte
;-------------------------------------------------------------------------
;input: a0 = byte
;
putchar:				;wait until tx buffer empty
	ldi r30, UCSR0A
	clr r31
	ld r29, Z
	sbrs	r29, UDRE0
;	sbis	UCSR0A, UDRE0
	rjmp	putchar
	ldi r30, UDR0
	clr r31
	st Z, a0
;	out	UDR0, a0
	ret
;-------------------------------------------------------------------------
