#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#ifdef MODE_LCD
	#include "lcd.h"
#endif

//mega168
#define UDR UDR0
#define UCSRA UCSR0A
#define UCSRB UCSR0B
#define UCSRC UCSR0C
#define UBRRL UBRR0L
#define UBRRH UBRR0H
#define UDRE UDRE0
#define RXEN RXEN0
#define TXEN TXEN0
#define RXCIE RXCIE0
#define OCR2 OCR2A
#define TCCR0 TCCR0B
#define TCCR2 TCCR2B

#define PAGESIZEB 128

#ifdef __AVR_ATmega168__
	#define ROMSIZEB 16384
	#define BOOTSIZEB 2048
#endif

#ifdef __AVR_ATmega88__
	#define ROMSIZEB 8192
	#define BOOTSIZEB 2048
#endif

#ifndef F_CPU
	#define F_CPU 8000000
#endif

// UART
#define UART_BAUD_RATE 19200

unsigned char buf[PAGESIZEB];



static void sendchar(unsigned char c) {
	while ((UCSRA & (1<<UDRE0)) == 0); //UDRE0
	UDR = c;
}

static unsigned char getchar(void) {
	while ((UCSRA & (1<<RXC0)) == 0);
	return UDR;
}


static void eraseflash(void) {
	volatile unsigned char i;
	unsigned int k;
	asm volatile ( "clr r30" );
	asm volatile ( "clr r31" );
	for (i=0; i<(ROMSIZEB-BOOTSIZEB)/128; i++) {
		while ((SPMCSR & (1<<SELFPRGEN)) == 1);
		SPMCSR = (1<<PGERS) | (1<<SELFPRGEN);
		asm volatile ( "spm" );
		while ((SPMCSR & (1<<SELFPRGEN)) == 1);
		SPMCSR = (1<<RWWSRE) | (1<<SELFPRGEN);
		asm volatile ( "spm" );
		asm volatile ( "ldi r24, 128" );
		asm volatile ( "add r30, r24" );
		asm volatile ( "adc r31, __zero_reg__" );
	}
	asm volatile ( "clr r30" );
	asm volatile ( "clr r31" );
	for (k=0; k<ROMSIZEB-BOOTSIZEB; k++) {
		asm volatile ( "lpm r0, Z+" );
		asm volatile ( "std Y+4, r0" );
		if (i != 255) {
#ifdef MODE_LCD
		lcd_string(DISP_ERROR2, 1);
#endif
			sendchar('0');
			return;
		}
	}
}


#if defined(MODE_OPT)

static void readflashline(void) {
	unsigned char i=0;
	asm volatile ( "in r30, 0x2a" );
	asm volatile ( "in r31, 0x2b" );
	for (i=0; i<PAGESIZEB; i++) {
		asm volatile ( "lpm r0, Z+" );
		asm volatile ( "add r1, r0" );
		asm volatile ( "inc r1" );
		while ((UCSRA & (1<<UDRE0)) == 0);
		asm volatile ( "sts 0xc6, r0" );
	}
	asm volatile ( "out 0x2a, r30" );
	asm volatile ( "out 0x2b, r31" );
	while ((UCSRA & (1<<UDRE0)) == 0);
	asm volatile ( "sts 0xc6, r1" );
	asm volatile ( "clr r1" );
}

static void writeflashline(void) {
	volatile unsigned char temp, temp2;
	unsigned char i=0, chksum=0;
	for (i=0; i<PAGESIZEB; i++) {
		temp = getchar();
		chksum += temp+1;
		buf[i] = temp;
	}
	temp = getchar();
	if (chksum != temp) {
#ifdef MODE_LCD
		lcd_string(DISP_ERROR, 1);
#endif
		sendchar('0');
		return;
	}
/*#ifdef MODE_LCD
	lcd_setcur(0,0);
	for (i=0; i<12; i++)
		lcd_hex_u08(buf[i]);
#endif*/
	for (i=0; i<PAGESIZEB; i+=2) {
		temp = buf[i];
		asm volatile ( "mov r0, r24 " );
		temp2 = buf[i+1];
		asm volatile ( "mov r1, r24 " );
		asm volatile ( "in r30, 0x2a" );
		asm volatile ( "in r31, 0x2b" );
		asm volatile ( "Wait_spm:" );
		asm volatile ( "in r24, 0x37" );
		asm volatile ( "sbrc r24, 0" );
		asm volatile ( "rjmp Wait_spm" );
		SPMCSR = (1<<SELFPRGEN);
		asm volatile ( "spm" );
		asm volatile ( "adiw r30, 2" );
		asm volatile ( "out 0x2a, r30" );
		asm volatile ( "out 0x2b, r31" );
	}
	asm volatile ( "clr __zero_reg__ " );


	asm volatile ( "sbiw r30, 63" );
	asm volatile ( "sbiw r30, 63" );
	asm volatile ( "sbiw r30, 2" );

	while ((SPMCSR & (1<<SELFPRGEN)) == 1);
	SPMCSR = (1<<PGWRT) | (1<<SELFPRGEN);
	asm volatile ( "spm" );

	while ((SPMCSR & (1<<SELFPRGEN)) == 1);
	SPMCSR = (1<<RWWSRE) | (1<<SELFPRGEN);
	asm volatile ( "spm" );

	while ((SPMCSR & (1<<SELFPRGEN)) == 1);
	for (i=0; i<PAGESIZEB; i++) {
		asm volatile ( "lpm r0, Z+" );
		asm volatile ( "std Y+4, r0" );
		asm volatile ( "push r30" );
		asm volatile ( "push r31" );
		temp = buf[i];
		asm volatile ( "pop r31" );
		asm volatile ( "pop r30" );
/*#ifdef MODE_LCD
		if (i<12) {
			lcd_setcur(i<<1,0);
			lcd_hex_u08(temp);
			lcd_setcur(i<<1,1);
			lcd_hex_u08(temp2);
		}
#endif*/
		if (temp != temp2) {
#ifdef MODE_LCD
			lcd_string(DISP_ERROR2, 1);
#endif
			sendchar('0');
			return;
		}
	}
}

#else

static void readflashline(void) {
	unsigned char i=0;
	for (i=0; i<PAGESIZEB; i++) {
		asm volatile ( "in r30, 0x2a" );
		asm volatile ( "in r31, 0x2b" );
		asm volatile ( "lpm __tmp_reg__, Z+" );
		asm volatile ( "add __zero_reg__, __tmp_reg__" );
		asm volatile ( "inc __zero_reg__" );
		asm volatile ( "out 0x2a, r30" );
		asm volatile ( "out 0x2b, r31" );
		while ((UCSRA & (1<<UDRE0)) == 0);
		asm volatile ( "sts 0xc6, __tmp_reg__" );
	}
	while ((UCSRA & (1<<UDRE0)) == 0);
	asm volatile ( "sts 0xc6, __zero_reg__" );
	asm volatile ( "clr __zero_reg__" );
}

static void writeflashline(void) {
	unsigned char i=0, chksum=0, temp, temp2;
	for (i=0; i<PAGESIZEB/2; i++) {
		temp = getchar();
		chksum += temp+1;
		temp2 = getchar();
		chksum += temp2+1;
		asm volatile ( "ldd __tmp_reg__, Y+3 " );
		asm volatile ( "ldd __zero_reg__, Y+4 " );
		asm volatile ( "in r30, 0x2a" );
		asm volatile ( "in r31, 0x2b" );
		SPMCSR = 1;
		asm volatile ( "spm" );
		asm volatile ( "adiw r30, 2" );
		asm volatile ( "out 0x2a, r30" );
		asm volatile ( "out 0x2b, r31" );
		asm volatile ( "clr __zero_reg__ " );
	}
	while ((UCSRA & (1<<RXC0)) == 0);
	temp = UDR;
	if (chksum != temp) {
		sendchar('0');
#ifdef MODE_LCD
		lcd_string(DISP_ERROR, 1);
#endif
		return;
	}

	// execute flash erase
	asm volatile ( "in r30, 0x2a" );
	asm volatile ( "in r31, 0x2b" );
	asm volatile ( "sbiw r30, 63" );
	asm volatile ( "sbiw r30, 63" );
	asm volatile ( "sbiw r30, 2" );
	SPMCSR = 3;
	asm volatile ( "spm" );
	while ((SPMCSR & (1<<PGERS)) == 1);

	// execute flash write
	SPMCSR = 5;
	asm volatile ( "spm" );
	while ((SPMCSR & (1<<PGWRT)) == 1);
}

#endif


void main(void) {
	unsigned int k;
	volatile unsigned char temp, temp2;
#ifdef MODE_LCD
	unsigned int cnt=0;
#endif

	DDRD = 0x02;	// Input-Port, auer TxD
	PORTD = 0xFF;	// Pull-Ups aktivieren

	k = 2;
	for (;;)
	{
		k--;
		if (k==0) break;
		if ((PIND & 0xFC) != 0xFC) {
#ifdef MODE_LCD
			DDRC = 0x3F;	// LCD
			PORTC = 0x00;
			lcd_init();
			lcd_string(DISP_MAIN, 0);
#endif

			// UART
			// Hilfsmakro zur UBRR-Berechnung ("Formel" laut Datenblatt)
			#define UART_UBRR_CALC(BAUD_,FREQ_) ((FREQ_)/((BAUD_)*16L)-1)
			UBRRH = (uint8_t)(UART_UBRR_CALC(UART_BAUD_RATE,F_CPU)>>8);
			UBRRL = (uint8_t)UART_UBRR_CALC(UART_BAUD_RATE,F_CPU);
			UCSRB = (1<<RXEN)|(1<<TXEN)|(1<<RXCIE);//|(1<<UDRIE);
			asm volatile ( "push r31" );
			asm volatile ( "lds r31, 0xc6" );
			asm volatile ( "pop r31" );

			asm volatile ( "out 0x2a, __zero_reg__" );
			asm volatile ( "out 0x2b, __zero_reg__" );

			temp = getchar();
			if (temp == 13) {
				sendchar(13);
#ifdef MODE_LCD
				lcd_string(DISP_CONN, 1);
#endif
				while(1) {
					temp = getchar();
					if (temp == 'w') {
#ifdef MODE_LCD
						lcd_string(DISP_WRITE, 1);
						lcd_setcur(14, 1);
						lcd_hex_u16(cnt++);
#endif
						sendchar('1');
						writeflashline();
						sendchar('1');
					}
					else if (temp == 'r') {
#ifdef MODE_LCD
						lcd_string(DISP_READ, 1);
						lcd_setcur(14, 1);
						lcd_hex_u16(cnt++);
#endif
						readflashline();
					}
					else if (temp == 'e') {
#ifdef MODE_LCD
						lcd_string(DISP_ERASE, 1);
#endif
						eraseflash();
						sendchar('1');
					}
					else if (temp == 'n') {
						asm volatile ( "out 0x2a, __zero_reg__" );
						asm volatile ( "out 0x2b, __zero_reg__" );
#ifdef MODE_LCD
						cnt = 0;
						lcd_clear();
						lcd_string(DISP_MAIN, 0);
#endif
						sendchar('1');
					}
					else if (temp == 'q') {
						sendchar('1');
#ifdef MODE_LCD
						lcd_clear();
#endif
						break;
					}
					else if (temp == 'i') {
						sendchar(0x1e);
						sendchar(0x94);
						sendchar(0x06);
					}
					else if (temp == 's') {
						temp = getchar();
						asm volatile ( "out 0x2a, r24" );
						temp2 = getchar();
						asm volatile ( "out 0x2b, r24" );
#ifdef MODE_LCD
						cnt = (temp2<<8) | temp;
#endif
						sendchar('1');
					}
					else if (temp == 13) {
						sendchar(13);
					}
				}
			}
			break;
		}
	}
	asm volatile ( "jmp 0x0000" );
}
