// mload.cpp : Definiert den Einsprungpunkt fr die Konsolenanwendung.
//


// INCLUDES

#include "stdafx.h"
#include "stdio.h"
#include "Serial.h"



// DEFINES

#define PAGESIZEB 128



// VARIABLEN

// Buffer fr HEX-Filedaten
char buf[32*1024], buf2[PAGESIZEB+1], chksum;
int maxadr=0, gl_error=0;



// PROGRAMM

unsigned char nibconv(char c)
{
	if (c >= '0' && c <= '9') return (unsigned char)c-'0';
	else if (c >= 'A' && c <= 'F') return (unsigned char)c+10-'A';
	else if (c >= 'a' && c <= 'f') return (unsigned char)c+10-'a';
	else {
		printf("(E) Error parsing HEX number.\n");
		gl_error = 1;
		return 0;
	}
}


// ######################### hexconv ################################
// Umrechnung Hex -> Byte
unsigned char hexconv(char c1, char c2)
{
	return (nibconv(c1) << 4) | nibconv(c2);
}


// ######################### hexnibble ################################
// Umrechnung Hexnibble -> Zahl
unsigned int hexnibble(char d)
{
	if (d >= '0' && d <= '9')
		return d-'0';
	else if (d >= 'A' && d <= 'F')
		return d-'A'+10;
	else if (d >= 'a' && d <= 'f')
		return d-'a'+10;
	return 0;
}


// ######################### hex2int ################################
// Umrechnung Hexstring -> Zahl
unsigned int hex2int(char *s)
{
	unsigned int i=(unsigned int)strlen(s), erg=0, c=0;
	while (i > 0)
	{
		erg |= hexnibble(s[i-1]) << (c*4);
		i--;
		c++;
	}
	return erg;
}


// ######################### numberParser ################################
// String in int umwandeln
unsigned int numberParser(char *zahl)
{
	if (zahl[0] == '0' && (zahl[1] == 'x' || zahl[1] == 'X'))
		return hex2int(&(zahl[2]));
	else
		return atoi(zahl);
}


char fgetchar(FILE *f) {
	if (!feof(f))
		return fgetc(f);
	else
		return 0;
}


unsigned char fgethex(FILE *f) {
	char a, b;
	unsigned char e;
	a = fgetchar(f);
	b = fgetchar(f);
	e = hexconv(a, b);
	chksum += e;
	return e;
}


unsigned char parseline(FILE *f) 
{
	int len, adr, t, i, extadr=0;
	chksum = 0;
	len = fgethex(f);
	adr = fgethex(f) << 8;
	adr |= fgethex(f);
	t = fgethex(f);
	switch (t) {
	case 0:
		for (i=0; i<len && gl_error==0; i++)
			buf[extadr+adr++] = fgethex(f);
		break;
	case 1:
		for (i=0; i<len && gl_error==0; i++)
			fgethex(f);
		break;
	case 4:
		extadr = fgethex(f)<<24;
		extadr |= fgethex(f)<<16;
		break;
	default:
		for (i=0; i<len && gl_error==0; i++)
			fgethex(f);
	}
	if (adr > maxadr)
		maxadr = adr;
	return (unsigned char)chksum + fgethex(f);
}


int main(int argc, char* argv[])
{
	char s[257], c;
	long i, k, len, com=1, br=115200;
	CSerial *com1 = new CSerial();
	FILE *f;

	printf("mload - Bootloader for new ATmega48/88/168 v2.1  (c) Simon\n");
	if (argc <= 2) {
		printf("Usage: mload com1@19200 mrmidi.hex\n  (Replace com1, 19200 and mrmidi.hex with custom values)");
		return -1;
	}

	if (argv[1][0] == 'c')
		sscanf(argv[1], "com%u@%u", &com, &br);

	printf("(I) Loading HEX file... (Ctrl+C to abort)\n");
	f = fopen(argv[2], "r");
	if (f != NULL) {
		if (!feof(f))
			if (fgetc(f) != ':') {
				printf("    (E) No valid INTEL HEX file (type error).\n");
				return -1;
			}
		while (!feof(f)) {
			if (parseline(f) != 0) {
				printf("    (E) No valid INTEL HEX file (checksum error).\n");
				return -1;
			}
			while (!feof(f)) {
				c = fgetc(f);
				if (c == ':')
					break;
			}
		}
		fclose(f);
		printf("    (I) File OK.\n");
	}
	else {
		printf("    (E) File \"%s\" read problem.\n", argv[2]);
		return -1;
	}

	if (com1->Open( com, br, 8, NOPARITY, ONESTOPBIT))
	{
		com1->SetTimeout(100);

		printf("(I) Connecting... (Ctrl+C to abort)\n");
		i = 0;
		do {
			com1->WriteCommByte(13);
			i++;
		} while (com1->ReadData(s, 1) == 0 && i < 50);
		if (i == 50) {
			printf("    (E) No response (connect).\n");
			return -1;
		}
		printf("    (I) Connected.\n");

		printf("(I) Query info...\n");
		com1->WriteCommByte('i');
		len = com1->ReadData(s, 3);
		if (len == 0) {
			printf("    (E) No response (query info).\n");
			return -1;
		}
		else if (len != 3) {
			printf("    (E) Unexpected response (query info).\n");
			return -1;
		}
		len = (s[0]&0xff)<<16;
		len |= (s[1]&0xff)<<8;
		len |= s[2]&0xff;
		switch (len) {
		case 0x1e9406:
			printf("    (I) ATmega168 found.\n");
			break;
		case 0x1e930a:
			printf("    (I) ATmega88 found.\n");
			break;
		case 0x1e9205:
			printf("    (I) ATmega48 found.\n");
			break;
		default:
			printf("    (I) Unknown MCU found.\n");
		}

		printf("(I) Flash erase in progress.\n");
		com1->SetTimeout(2000);
		com1->WriteCommByte('e');
		if (com1->ReadData(s, 1) == 0) {
			printf("    (E) No response (erase).\n");
			return -1;
		}
		if (s[0] == '0') {
			printf("\n    (E) Error response (erase).\n");
			return -1;
		}
		else if (s[0] != '1') {
			printf("    (E) Wrong response (erase).\n");
			return -1;
		}
		printf("    (I) Flash blanked.\n");

		com1->SetTimeout(100);
		com1->WriteCommByte('n');
		if (com1->ReadData(s, 1) == 0) {
			printf("    (E) No response (setup write).\n");
			return -1;
		}
		if (s[0] == '0') {
			printf("\n    (E) Error response (setup write).\n");
			return -1;
		}
		else if (s[0] != '1') {
			printf("    (E) Wrong response (setup write).\n");
			return -1;
		}

		printf("(I) Flash update in progress.");
		i = 0;
		for (i=0; i<maxadr; i+=PAGESIZEB) {
			chksum = 0;
			for (k=0; k<PAGESIZEB; k++)
				chksum+=buf[i+k]+1;
			com1->WriteCommByte('w');
			if (com1->ReadData(s, 1) == 0) {
				printf("\n    (E) No response (write memory).\n");
				return -1;
			}
			if (s[0] == '0') {
				printf("\n    (E) Error response (write memory).\n");
				return -1;
			}
			else if (s[0] != '1') {
				printf("\n    (E) Wrong response (write memory).\n");
				return -1;
			}
			com1->SendData(&buf[i], PAGESIZEB);
			com1->WriteCommByte(chksum);
			printf(".");
			if (com1->ReadData(s, 1) == 0) {
				printf("\n    (E) No response (write memory).\n");
				return -1;
			}
			if (s[0] == '0') {
				printf("\n    (E) Error response (write memory).\n");
				return -1;
			}
			else if (s[0] != '1') {
				printf("\n    (E) Wrong response (write memory).\n");
				return -1;
			}
		}
		printf("\n    (I) Write OK.\n");

		com1->WriteCommByte('n');
		if (com1->ReadData(s, 1) == 0) {
			printf("    (E) No response (setup read).\n");
			return -1;
		}
		if (s[0] != '1') {
			printf("    (E) Wrong response (setup read).\n");
			return -1;
		}

		printf("(I) Verifying.");
		i = 0;
		for (i=0; i<maxadr; i+=PAGESIZEB) {
			chksum = 0;
			com1->WriteCommByte('r');
			len = com1->ReadData(buf2, PAGESIZEB+1);
			if (len == 0) {
				printf("\n    (E) No response (read memory).\n");
				return -1;
			}
			else if (len != PAGESIZEB+1) {
				printf("\n    (E) Transfer error (read memory).\n");
				return -1;
			}
			for (k=0; k<PAGESIZEB; k++) {
				chksum+=buf2[k]+1;
				if ((buf2[k]&0xff) != (buf[i+k]&0xff)) {
					printf("\n    (E) Memory differs from HEX file at 0x%.4X (MEM=0x%.2X, DATA=0x%.2X).\n", k+i, buf2[k]&0xff, buf[i+k]&0xff);
				}
			}
			if (chksum != buf2[PAGESIZEB]) {
				printf("\n    (E) Checksum error (read memory).\n");
				return -1;
			}
			printf(".");
		}
		printf("\n    (I) Memory OK.\n");
		printf("(I) Starting program.\n");
		com1->WriteCommByte('q');
		if (com1->ReadData(s, 1) == 0) {
			printf("    (E) No response from Mr.Midi (start app).\n");
			return -1;
		}
		if (s[0] != '1') {
			printf("    (E) Wrong response from Mr.Midi (start app).\n");
			return -1;
		}
		printf("(I) Update complete.\n");

		com1->Close();
	}
	else {
		printf("(E) Com%u open problem.\n", com);
		return -1;
	}

	return 0;
}

